#!/usr/bin/ruby -w

# Scotty TNG - ping/snmp based realtime network monitoring
#
# $Id: client.rb 143 2008-10-09 10:01:29Z liske $
#
# Authors:
#   Thomas Liske <liske@ibh.de>
#   Andre Ellguth <ellguth@ibh.de>
#
# Copyright Holder:
#   2008 (C) IBH IT-Service GmbH [http://www.ibh.de/]
#
# License:
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this package; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
#

require 'gtk2'
require 'cairo'
require 'libglade2'
require 'drb'
require 'drbfire'
require 'rinda/ring'
require 'rinda/tuplespace'
require 'zlib'
require 'rexml/document'

basedir = File.dirname(File.expand_path(__FILE__))

begin
    Dir.chdir('/usr/lib/scotty-tng/client')
rescue
end

require 'utils.rb'
require 'gtkqueue.rb'
require 'view.rb'
require 'node.rb'
require 'notify.rb'
require 'chart.rb'
require 'entities.rb'

begin
    Dir.chdir('/usr/share/scotty-tng/client')
rescue
end

class Client
  include GetText

  attr :glade

  @@actions = Hash.new
  @@commands = Hash.new
  @@transitions = Array.new

  def self.getActs
    @@actions
  end

  def self.getCmds
    @@commands
  end

  def self.getTrans
    @@transitions
  end

  def self.getTS
    @@ts
  end

  def self.getSIcon
    @@sicon
  end

  def self.getTooltips
    @@tt
  end

  def initialize(basedir, path_or_data, root = nil, domain = nil, localedir = nil, flag = GladeXML::FILE)
    @@singleton = self

    @basedir = basedir
    bindtextdomain(domain, localedir, nil, "UTF-8")
    Logger.setClient(self)
    @glade = GladeXML.new(path_or_data, root, domain, localedir, flag) {|handler| method(handler)}

    @glade.widget_names.each{|wname|
	instance_variable_set("@#{wname.gsub(/-/,'_')}", @glade.get_widget(wname))
    }

    begin
	@@sicon = Gtk::StatusIcon.new
	@@sicon.file = "res/rsb2.png"
    rescue
	@@sicon = nil
    end

    @@tt = Gtk::Tooltips.new
    @@tt.enable
    @drbErrMux = Mutex.new

    ["ok", "info", "warning", "critical"].each{|l|
	tag = Gtk::TextTag.new(l)
	tag.foreground = Status.text_color(l)
	@textview_log.buffer.tag_table.add(tag)
    }

    # Show main window and startup splash
    Gtk::Window.auto_startup_notification = false
    @image_loading.set_file("res/scotty-tng.png")
    @window_main.show
    @window_loading.show
    Gtk::Window.auto_startup_notification = true

    load_config("client.conf")
  end

  # Process pending Gtk events and return
  def iter
    while Gtk.events_pending? do
	Gtk.main_iteration
    end
  end

  # Print a loging message on screen
  def logtxt(text, mayiter=true, tag=nil, time=Time.now)
    if @textview_log.buffer.char_count > 0 then
	line = "\n"
    else
	line = ""
    end
    text = text.to_s.gsub(/\n+/, ' ')
    line += "[#{Time.at(time).strftime('%X')}] #{text}"

    if tag
	@textview_log.buffer.insert(@textview_log.buffer.end_iter, line, tag)
    else
	@textview_log.buffer.insert(@textview_log.buffer.end_iter, line)
    end

    iter if mayiter

    @scrolledwindow_log.vadjustment.value = @scrolledwindow_log.vadjustment.upper -
	@scrolledwindow_log.vadjustment.page_size

    iter if mayiter
  end

  def load_config(fn)
    Notify.init

    iter

    logtxt("Loading '#{fn}'...")
    xml = REXML::Document.new(File.open(fn))

    iter

    xml.elements.each("/config/commands/*") { |xcmd|
	@@commands[xcmd.name] = xcmd.to_s.sub(/^<#{xcmd.name}>/, "").sub(/<\/#{xcmd.name}>$/, "")
    }

    iter

    xml.elements.each("/config/transitions/transit") { |xtrans|
	@@transitions.push([xtrans.attributes["from"], xtrans.attributes["to"], xtrans.attributes["type"], xtrans.attributes["class"], xtrans.attributes["action"]])
    }

    iter

    xml.elements.each("/config/actions/action") { |xact|
	cmds = Array.new
	xact.elements.each { |xc|
	    attrs = Hash.new
	    xc.attributes.each{ |name, value|
		attrs[name] = value
	    }
	    cmds.push([xc.name, attrs])
	}
	@@actions[xact.attributes["name"]] = [cmds, xact.attributes["interval"].to_i]
    }

    xml.elements.each("/config/gui/status") { |xstatus|
	Status.set_color(xstatus.attributes["name"], xstatus.elements["color"].text) if xstatus.elements["color"]
	Status.set_icon(xstatus.attributes["name"], xstatus.elements["icon"].text) if xstatus.elements["icon"]
    }
  end

  # Initialize the DRb stuff
  def init_drb
    iter

    logtxt("Starting DRb service")

    uri = ARGV.pop

    begin
	DRb.start_service uri, nil, {DRbFire::ROLE => DRbFire::CLIENT}

	if uri then
	    logtxt("Using #{uri}")
	    @@ts = Rinda::TupleSpaceProxy.new DRbObject.new_with_uri(uri)
	else
	    rf = Rinda::RingFinger.new
	    doloop = 1

	    while doloop == 1
		logtxt("Lookup ring servers")
		@rs = Hash.new
		rf.lookup_ring { |s|
		    @rs[s.__drburi] = s
		}

		if @rs.size == 0 then
		    logtxt("Error: Could not find any ring server!")
		    dialog = Gtk::Dialog.new("Error",
				 @window_loading,
				 Gtk::Dialog::MODAL,
				 [Gtk::Stock::REFRESH, Gtk::Dialog::RESPONSE_ACCEPT],
				 [Gtk::Stock::CANCEL, Gtk::Dialog::RESPONSE_REJECT]
		    )
		    dialog.vbox.add(Gtk::Label.new("Could not find ring server!"))
		    dialog.show_all

		    if dialog.run == Gtk::Dialog::RESPONSE_ACCEPT then
			dialog.destroy
			next
		    end

		    exit 0
		elsif @rs.size > 1 then
		    @dialog_selring.set_modal(true)
		    @dialog_selring.show_all
		    exit if @dialog_selring.run != Gtk::Dialog::RESPONSE_ACCEPT

		    doloop = 0
		else
		    @@ts = @rs.values.shift
		    logtxt("Using #{@@ts.__drburi}")

		    doloop = 0
		end
	    end
	end

	iter

	@viewsets = @@ts.read_all([:ScottyTNGd, nil, nil])

	if @viewsets.size > 1 then
	    @dialog_selview.show_all
	    exit if @dialog_selview.run != Gtk::Dialog::RESPONSE_ACCEPT
	elsif @viewsets.size == 1 then
	    @view = @viewsets[0][2]
	    @window_main.title = "Scotty TNG [#{@viewsets[0][1]}]"
	else
	    drbError("No views found!")
	end

	# Start observer and load views.
	startObserver

	xml = REXML::Document.new( Zlib::Inflate.inflate(@view) )
	total = REXML::XPath.first(xml, "count(/views/view/*)")

	@progressbar_loading.pulse_step = 0.02
	@progressbar_loading.text = "Loading nodes..."
	nodes = @@ts.read_all([:Node, nil, nil])
	nodes.each{|n|
	    @progressbar_loading.pulse
	    iter

	    Node.add(n[1], nil, n[2])
	}

	@progressbar_loading.text = "Loading charts..."
	charts = @@ts.read_all([:DRingLOpts, nil, nil, nil])
	charts.each{|c|
	    @progressbar_loading.pulse
	    iter

	    Chart.add(c[1], c[2], nil, c[3])
	}

	@progressbar_loading.fraction = 0
	prog = 0
	xml.elements.each("/views/view") { |xview|
	    @progressbar_loading.text = "Loading view '" + xview.attributes["title"] + "'..."

	    View.new(@notebook_main, xview){
		prog = prog + 1
		@progressbar_loading.fraction = prog.to_f/total.to_f

		iter
	    }
	}

	@window_loading.hide
    rescue DRb::DRbError, Errno::ECONNREFUSED, Errno::ENETUNREACH, Errno::EHOSTUNREACH
	drbError($!)
    end

    return 1
  end

  # This method is called by the observer threads if a DRb::DRbError has occured. We
  # use a mutex to show errors only once and terminate/restart the client.
  def drbError(msg)
	return unless @drbErrMux.try_lock

	Gtk.queue do
	    dialog = Gtk::MessageDialog.new(
		@window_main, Gtk::Dialog::DESTROY_WITH_PARENT,
		Gtk::MessageDialog::ERROR, Gtk::MessageDialog::BUTTONS_YES_NO,
		"Communication problem: #{msg}\n\nShould Scotty TNG be restarted to recover?")

	    dialog.run{|resp|
		case resp
		    when Gtk::Dialog::RESPONSE_YES
			# Restore path and start client again.
			Dir.chdir(@basedir)
			exec("/usr/bin/ruby", __FILE__)
		    else
			exit
		    end
	    }
	end
  end

  # Start observers of changes in the TupleSpace as separate threads.
  def startObserver
    # Record graph updates
    @obsthread = Thread.start do
	begin
	    observer = @@ts.notify 'write', [:DRingLVals, nil, nil, nil]
	    observer.each{|e, t|
		nname = t[1]
		sname = t[2]
		dval = t[3]

		chart = Chart.getChart([nname, sname])
		if chart then
		    Gtk.queue do
			chart.pushVal(dval)
		    end
		end
	    }
	rescue DRb::DRbError
	    drbError($!)
	end
    end

    # Record sensor events
    @ostatsthread = Thread.start do
	begin
	    obsstatus = @@ts.notify 'write', [:SensorStatus, nil, nil, nil]
	    obsstatus.each{|e, t|
		nname = t[1]
		sname = t[2]
		status = t[3]
		chart = Chart.getChart([nname, sname])
		if chart then
		    node = Node.getNode(nname)
		    Gtk.queue do
			chart.setStatus(status)
			node.setStatus(sname, status) if node
		    end
		end
	    }
	rescue DRb::DRbError
	    drbError($!)
	end
    end

    # Record nagios events
    @onagiosthread = Thread.start do
	begin
	    obsnagios = @@ts.notify 'write', [:NagiosStatus, nil, nil, nil]
	    obsnagios.each{|e, t|
		name = t[1]
		service = t[2]
		status = t[3]
		node = Node.getNode(name)
		if node then
		    Gtk.queue do
			node.setNagiosStatus(service, status)
		    end
		end
	    }
	rescue DRb::DRbError
	    drbError($!)
	end
    end
  end


  ##
  ## window_main
  ##

  # set urgency hint
  def self.setUrgency(hint)
    @@singleton.setUrgency(hint)
  end

  def setUrgency(hint)
    @window_main.urgency_hint = hint
  end

  # Set cairo context color by string
  def self.set_ct_scol(ct, color)
    begin
        ct.set_source_color(color)
    rescue
	case color.downcase
	    when "red"
		ct.set_source_rgb(1, 0, 0)
	    when "darkred"
		ct.set_source_rgb(0.7, 0, 0)
	    when "green"
		ct.set_source_rgb(0, 1, 0)
	    when "darkgreen"
		ct.set_source_rgb(0, 0.7, 0)
	    when "blue"
		ct.set_source_rgb(0, 0, 1)
	    when "darkblue"
		ct.set_source_rgb(0, 0, 0.7)
	    when "orange"
		ct.set_source_rgb(1, 0.65, 0)
	    when "darkorange"
		ct.set_source_rgb(1, 0.55, 0)
	    else
		# Nothing matched, black is always good!
		ct.set_source_rgb(0, 0, 0)
	end
    end
  end

  # Formats a value for a given format (=[t]ime, [p]ercent, [b]ytes...)
  def self.format_data(format, val)
    case format
	# Time in ms
	when "t"
	    return sprintf("%dms", val) if val < 1000
	    return sprintf("%ds", val/1000) if val < 60000
	    return sprintf("%dm", val/60000) if val < 3600000
	    return sprintf("%dh", val/3600000)
	# Percent
	when "p"
	    return sprintf("%d%%", val)
	# Bitrate
	when "r"
	    return sprintf("%.1fT", val / 1000000000000.0) if val >= 1000000000000
	    return sprintf("%.1fG", val / 1000000000.0   ) if val >= 1000000000
	    return sprintf("%.1fM", val / 1000000.0      ) if val >= 1000000
	    return sprintf("%.1fk", val / 1000.0         ) if val >= 1000
	    return sprintf("%d", val)
	when "b"
	    return "yes" if val > 0
	    return "no"
	when "ups_bat"
	    case val
	    when 2
		return "normal"
	    when 3
		return "low"
	    when 4
		return "depleted"
	    else
		return "unknown"
	    end
	when "ups_src"
	    case val
	    when 2
		return "none"
	    when 3
		return "normal"
	    when 4
		return "bypass"
	    when 5
		return "battery"
	    when 6
		return "booster"
	    when 7
		return "reducer"
	    else
		return "other"
	    end
	# Anything else
	else
	    return sprintf("%d", val)
    end
  end

  # Terminate if main window is closed
  def window_main_delete_event_cb(widget, arg0)
    Gtk.main_quit
  end

  def gtk_main_quit(widget)
    Gtk.main_quit
  end

  def togglebutton_mute_toggled_cb(widget)
    Notify.muteSound(widget.active?)

    if widget.active? then
	@image_mute.stock = Gtk::Stock::CANCEL
    else
	@image_mute.stock = Gtk::Stock::CLOSE
    end
  end

  def togglebutton_desktop_toggled_cb(widget)
    if widget.active? then
	@window_main.maximize
	@window_main.stick
	@window_main.hide
	@window_main.type_hint = Gdk::Window::TYPE_HINT_DESKTOP
	@window_main.show
	@image_desktop.stock = Gtk::Stock::FULLSCREEN
    else
	@window_main.hide
	@window_main.type_hint = Gdk::Window::TYPE_HINT_NORMAL
	@window_main.show
	@image_desktop.stock = Gtk::Stock::LEAVE_FULLSCREEN
    end
  end

  def button_about_clicked_cb(widget)
    @aboutdialog.run
    @aboutdialog.hide
  end

  ##
  ## dialog_selring
  ##

  def dialog_selring_show_cb(widget)
    liststore = Gtk::ListStore.new(String)

    renderer = Gtk::CellRendererText.new
    col = Gtk::TreeViewColumn.new("RingServer", renderer)
    col.set_cell_data_func(renderer) do |col, renderer, model, iter|
	renderer.text = iter[0]
    end

    @treeview_ring.selection.mode = Gtk::SELECTION_SINGLE
    @treeview_ring.model = liststore
    @treeview_ring.append_column(col)

    @rs.values.each{|views|
      viewset = liststore.append
      viewset[0] = views.__drburi
    }
  end


  ##
  ## dialog_selview
  ##
  def dialog_selview_show_cb(widget)
    treestore = Gtk::TreeStore.new(String, REXML::Document)

    renderer = Gtk::CellRendererText.new
    col = Gtk::TreeViewColumn.new("View", renderer)
    col.set_cell_data_func(renderer) do |col, renderer, model, iter|
	renderer.markup = iter[0]
    end

    @treeview_views.selection.mode = Gtk::SELECTION_SINGLE
    @treeview_views.model = treestore
    @treeview_views.append_column(col)

    @viewsets.each{|views|
	viewset = treestore.append(nil)
	viewset[0] = views[1]
	viewset[1] = views[2]
    }

    @treeview_views.expand_all
  end

  def treeview_views_row_activated_cb(widget, arg0, arg1)
    dialog = widget.toplevel
    case dialog.name
      when "dialog_selview"

      it = @treeview_views.model.get_iter(arg0)
      if it.get_value(1) != nil then
	dialog.hide
        iter

	@view = it.get_value(1)
	@window_main.title = "Scotty TNG [#{it.get_value(0)}]"

	dialog.response(Gtk::Dialog::RESPONSE_ACCEPT)
      end
      when "dialog_selring"

      it = @treeview_ring.model.get_iter(arg0)
      if it.get_value(0) != nil then
	@@ts = @rs[it.get_value(0)]
	dialog.hide
	dialog.response(Gtk::Dialog::RESPONSE_ACCEPT)
      end
      logtxt("Using #{@@ts.__drburi}")
      else
      return
    end
  end

  def button_exit_clicked_cb(widget)
    dialog = widget.toplevel
    dialog.hide
  end
end

Gtk.main_with_queue(250) if Client.new(basedir, "client.glade", nil, "Scotty TNG").init_drb
